#include "stdafx.h"
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>  // for compatibility with Visual C++
#include "Ygl.h"
#include <string.h>
#include <stdarg.h>
#include "vector.h"
#include "syntax.h"
#include "box.h"
#include "grid.h"
#include "field.h"
#include "table.h"
#include "peak.h"
#include "interpol.h"
#include "simulation.h"
#include "gate.h"
#include "fplot.h"

extern double CHARGE_LOSS;

extern double get_sim_time(TokenString &TS);
extern void setTortuosity(TokenString &TS, int p, double *Diff);
extern const char *getMethod(char *smethod, CaMethod &, BufMethod &);
extern void getRun( TokenString &TS, int i, bool *adaptive=0, char *PDEmethod=0, double *time=0, double *dt=0, 
             double *dtMax=0, double *accuracy=0, double *accuracy0=0, double *maxError=0, double *dtStretch=0, 
             int *n0=0, double *nStretch=0, int *nMax=0, double *ODEaccuracy=0);

//*******************************************************************************************

SimulationObj::SimulationObj(TokenString &TS) {

  initialize();

  int pos;
  char temp[512];

  totalSimTime = get_sim_time(TS); // make sure simulation time is computed before Plot Array is constructed!

  if (TS.token_count("geometry","spherical") || TS.token3_count("geometry",ASSIGN_TOKEN,"spherical")) 
      GEOMETRY = SPHERICAL; 
  else if  (TS.token2_count("geometry","conical") || TS.token3_count("geometry",ASSIGN_TOKEN,"conical"))
      GEOMETRY = CONICAL; 
  else if  (TS.token2_count("geometry","cylindrical") || TS.token3_count("geometry",ASSIGN_TOKEN,"cylindrical"))
      GEOMETRY = CYLINDRICAL;
  else if (TS.token2_count("geometry","cartesian.3D") || TS.token3_count("geometry",ASSIGN_TOKEN,"cartesian.3D"));
  else if (TS.token_count("geometry"))  TS.errorMessage( TS.token_index("geometry") + 2, StrCpy("Unknown geometry specifier"));
  else GEOMETRY = CARTESIAN3D;

  Params  = &TS;
  BCArray = new BCarrayObj(TS);
  Synapse = new RegionObj(TS);
  Grid    = new GridObj( *Synapse, TS);
  Synapse->bindToGrid( *Grid );
  FieldObj::setStaticData(*Synapse, *Grid, *BCArray); 
  Ca      = new FieldObj(TS);
  if (!Ca->source_num) throw StrCpy("No calcium channels specified");
  Buffers = new BufferArray(TS);
  initTortuosity();
  Gates   = new KineticObj(this);
  if ( TS.token_count("Import", &pos) ) Import( TS.line_string( pos + 1, temp ) );
  Plots   = new PlotArray(*this);
  Plots->draw_all();
  strcpy(m_PDEscheme, "default");

  for (int i=1; i <= TS.token_count("Run"); i++) // make sure all "run" and "current" statements are Ok
  try { 
     bool adaptive; double t;
     getRun( *Params, i, &adaptive, m_PDEscheme, &t, &m_dt0, &m_dtMax, &m_accuracy, &m_accuracy0, &m_maxError, 
                                                    &m_dtStretch, &m_n0, &m_nStretch, &m_nMax, &m_ODEaccuracy);   
     getMethod(m_PDEscheme, CaStep, BufStep );
     int pos = TS.token_index("current", i) + 1;
     if (TS.equal(pos, VAR_TOKEN) || TS.equal(pos, ASSIGN_TOKEN) ) pos ++;
     ExpressionObj(TS, pos, "Bad expression for the current", this).Evaluate();
  } catch (char *str) {
     TS.errorMessage( TS.token_index("Run", i), str, "Cannot process the \"Run\"/\"current\" statement");
  }

  m_dt0 = 1.0e-3; m_dtMax = 0.1; m_ODEaccuracy = m_accuracy = 1e-4; m_accuracy0 = 1e-3;
  m_dtStretch = 1.05; m_n0 = 2; m_nStretch = 1.2; m_nMax = 22; m_reduce = 0.8; m_maxError = 0.05;
  strcpy(m_PDEscheme, "default");

  Params->get_param("adaptive.dt0",0,&m_dt0);             Params->get_param("adaptive.dtMax",0,&m_dtMax);
  Params->get_param("adaptive.accuracy0",0,&m_accuracy0); Params->get_param("adaptive.accuracy", 0,&m_accuracy); 
  Params->get_param("adaptive.dtStretch",0,&m_dtStretch); Params->get_int_param("adaptive.n0", 0,&m_n0); 
  Params->get_param("adaptive.nStretch",0,&m_nStretch);   Params->get_int_param("adaptive.nMax", 0,&m_nMax);
  Params->get_param("adaptive.maxError",0,&m_maxError);   Params->get_param("adaptive.reduce", 0,&m_reduce); 
  Params->get_param("ODE.accuracy",0,&m_ODEaccuracy);     Params->get_string_param("PDE.scheme", m_PDEscheme);

 }

//*******************************************************************************************

SimulationObj::~SimulationObj() {

  if (Plots)          delete Plots;
  if (Gates)          delete Gates;
  if (DiffArray)      killTortuosity();
  FieldObj::kill_tridiag();      
  if (Buffers)        delete Buffers;
  if (Ca)             delete Ca;
  if (Grid)           delete Grid;
  if (Synapse)        delete Synapse;
  if (BCArray)        delete BCArray;
}

//********************************************************************************************
//********************************************************************************************


void SimulationObj::killTortuosity() {
  if (DiffArray) {
     for (int ti = 0; ti < DiffNum; ti ++) delete [] DiffArray[ti]; 
     delete [] DiffArray;
  }
}

//*****************************************************************************

void SimulationObj::initTortuosity() {

  bool undefined = false;
  int  bi;
  double *Diff = 0;
  long l = 0;

  DiffNum = 0;
  if ( Ca->tortDefined ) DiffNum++; else undefined = true;

  for (bi = 0; bi < Buffers->buf_num; bi++) 
    if ( Buffers->array[bi]->tortDefined ) DiffNum++;
    else undefined = true;

  if ( undefined ) DiffNum ++;

  DiffArray = new double *[DiffNum];
  int ind = 0;

  if ( undefined ) {   // at least one field has undefined tortuosity
    Diff = new double[ FieldObj::Size ];
    //fprintf(stdlog,"\n Diff -> %ld\n", Diff);
    DiffArray[ind++] = Diff; 
    if ( !Params->token_count("tortuosity") && !Params->token_count("all.tortuosity") ) 
      for (l = 0; l < FieldObj::Size; l++) Diff[l] = 1.0;
    else {
      if (VERBOSE) fprintf(stdlog,"\n### Global (default) tortuosity function specified:\n");
      int p = 0;
      if ( Params->token_count("tortuosity", &p) ) { if (Params->equal(p + 1, "=")) p ++; }
      else if ( Params->token_count("all.tortuosity", &p) ) { if (Params->equal(p + 1, "=")) p ++; }
      setTortuosity(*Params, p+1, Diff);
    }
  }

  if ( Ca->tortDefined ) DiffArray[ind++] = Ca->Diff; else Ca->Diff = Diff;
  for (bi = 0; bi < Buffers->buf_num; bi++) 
    if ( Buffers->array[bi]->tortDefined ) DiffArray[ind++] = Buffers->array[bi]->Diff; 
       else Buffers->array[bi]->Diff = Diff;

  // fprintf(stdlog,"\n ind=%d defined=%d \n", ind, defined);
}
//*****************************************************************************

double *SimulationObj::ResolveID(const char *svar, double **t) {

  if (VERBOSE > 8) fprintf(stdlog, ">> ResolveID: Resolving %s\n", svar);

    if (Ca) {
      //     if ( equal(svar, Ca->ID) )  
      // { if (t) *t = &Ca->Time; return (*Ca)(); }
     if ( equal(svar, "ICa") )  
       { if (t) *t = &Ca->Time; return &(Ca->ICa); }
     if ( equal(svar, "Charge.loss") )  
       { if (t) *t = &Ca->Time; return &(CHARGE_LOSS); }
    }     

    //    if (Buffers)
    //      for (j = 0; j < Buffers->buf_num; j++)
    //	if ( equal(svar,Buffers->array[j]->ID) )  
    //          { if (t) *t = &Buffers->array[0]->Time;  return (*Buffers->array[j])(); }

    if (Gates) return Gates->ResolveID(svar, t);

    return 0;
}

//**************************************************************************


const char *SimulationObj::ResolvePtr(double *ptr) {

    if (ptr == &CHARGE_LOSS) return "Charge.loss";

    if (Ca) {
       if ( ptr == Ca->elem )  return Ca->ID; 
       else if ( ptr == &(Ca->ICa) )  return "ICa";
       else if ( ptr == &CHARGE_LOSS )  return "Charge.loss";
    } 

    if (Buffers)
      for (int j = 0; j < Buffers->buf_num; j++)
	if ( ptr == Buffers->array[j]->elem )  return  Buffers->array[j]->ID; 

    if (Gates) return Gates->ResolvePtr(ptr);
    return 0;
    //fprintf(stdlog,"ResolvePtr: cannot resolve pointer %X\n", ptr );
    //exit(1);
}

//***************************************************************************

FieldObj *SimulationObj::ResolveField(const char *label) {
      FieldObj *field = 0;
      if (Ca) {  if ( equal(label, Ca->ID) ) field = Ca; }
      if (Buffers) for (int i=0; i < Buffers->buf_num; i++)
        if ( equal(label, (Buffers->array[i])->ID) ) field = Buffers->array[i];
      return field;
}


//**************************************************************************************************


void SimulationObj::Export(const char *filename)
{
  FILE *f;
  char endtag[7];
  strcpy(endtag,"endtag");
 
  if (VERBOSE) 
    fprintf(stdlog,"\n#### Dumping all state variables into file %s\n", filename);

  f = fopenAssure(filename, "w", "Simulation state export", "");

  if (Ca) {
    fwrite( (void *)&(Ca->size), sizeof(long), 1, f);
    fwrite( (void *)(Ca->elem), sizeof(double), Ca->size, f);
  }

  if (Buffers) 
    for (int i=0; i< Buffers->buf_num; i++) 
      fwrite( (void *)(Buffers->array[i]->elem), sizeof(double), Buffers->array[i]->size, f);

  if (Gates) {
     unsigned short int m = Gates->var_num;
     fwrite( &m, sizeof(unsigned short int), 1, f );
     fwrite( (void *)(Gates->var->elem), sizeof(double), Gates->var_num, f);
  }

  fwrite( (void *)endtag, sizeof(char), strlen(endtag)+1, f);
  fclose(f);
}


//**************************************************************************************************

void SimulationObj::Import(const char *filename)
{
  FILE *f;
  char endtag[7];
 
  if (VERBOSE) 
    fprintf(stdlog,"\n#### Importing all state variables from file %s\n", filename);

  f = fopenAssure(filename, "r", "Simulation state import", "");

  long n;
  long Size;

  if (Ca) {
    Size = Ca->size;
    fread( (void *)&n, sizeof(long), 1, f);
    if (n != Size) 
      throw makeMessage("Can't import the concentration fields: wrong number of elements (%ld vs %ld)\n", n, Size);
    fread( (void *)(Ca->elem), sizeof(double), Size, f);
    // Ca->initialAve = Ca->average();
  }

  if (Buffers) 
    for (int i=0; i< Buffers->buf_num; i++) {
      fread( (void *)(Buffers->array[i]->elem), sizeof(double), Size, f);
      // Buffers->array[i]->initialAve = Buffers->array[i]->average();
    }

  if (Gates) {
    unsigned short int m = 0;
    fread( &m, sizeof(unsigned short int), 1, f);
    if (m != Gates->var_num)  
      throw makeMessage("Can't import the dynamic variables: wrong number of ODE variables (%d vs %d)\n", m, Gates->var_num);
    fread( (void *)(Gates->var->elem), sizeof(double), Gates->var_num, f);
    Gates->locations->reset();
    Gates->Evaluate();
  }

  fread( (void *)endtag, sizeof(char), strlen("endtag")+1, f );
  if ( !equal(endtag,"endtag") )  
    throw makeMessage("in Import: file size mismatch (endtag=%s)\n", endtag);

  fclose(f);
}

//**************************************************************************
//          H I N E S / N O N - A D A P T I V E    M E T H O D                 
//**************************************************************************

void SimulationObj::FixedTimeStep(double T, int n)
{
  VectorObj Next(FieldObj::Size);

  double dt = (T - Ca->Time) / double(n + 0.5);

  RunStatusString *rss = 0;
  if (VERBOSE > 4) rss = new RunStatusString(30,"time",&(Ca->Time),T);						       
  else if (VERBOSE > 1) rss = new RunStatusString(1,"time",&(Ca->Time),T);						       
  Ca->Step(CaStep, Next, *Buffers, 0.5 * dt, T);  // half a step for staggering
  
  for (int i = 1; i <= n; i++)
    {
    Buffers->Step(BufStep, Next, *Ca, dt, T);
    if (VERBOSE > 4) Gates->RungeKuttaAdaptive(dt, m_ODEaccuracy, 0, rss, T);
      else Gates->RungeKuttaAdaptive(dt, m_ODEaccuracy, 0, 0, T);
    Ca->Step(CaStep, Next, *Buffers, dt, T);
    if (VERBOSE > 1)  rss->update('.');
    Plots->draw_all(); 
    }

  Buffers->Step(BufStep, Next, *Ca, 0.5 * dt, T); // do half a time step for staggering
  Gates->RungeKuttaAdaptive(0.5 * dt, m_ODEaccuracy, 0, 0, T);
  Plots->draw_all();

  if (rss) delete rss;
}

//**************************************************************************
//                    F I N D   T I M E   S T E P
//**************************************************************************

void SimulationObj::SetTimeStep()
{
  VectorObj Next(FieldObj::Size); 
  FieldObj  Next2(*Ca); 
  double    old, half, full, error;

  if (VERBOSE) {
    fprintf(stdlog, " ## Setting the initial time step: {"); fflush(stdlog); }

  Ca->Step(CaStep, Next, *Buffers, m_dt0, -1); 
  half = Next.norm();

    do
      {
      old = half;
      m_dt0 *= 0.5;
      Ca->Step(CaStep, Next2, *Buffers, m_dt0, -1);
      half = Next2.norm();
      Next2.Time = Ca->Time + m_dt0;
      Next2.Step(CaStep, Next, *Buffers, m_dt0, -1);
      full = Next.norm();
      if (VERBOSE > 0 && VERBOSE < 8) { fprintf(stdlog, "."); fflush(stdlog); }
      else if (VERBOSE > 8)
         fprintf(stdlog, " > halving dt: dt = 2 * %g: res=%g  dt=%g: res=%g\n",m_dt0, full, 2*m_dt0,old);
      }
    while ( (error = fabs( 2 * (old - full) / (old + full + 1.0e-12) )) > m_accuracy0 );
    *Ca = Next2;
    // Ca.Time += dt;
    m_dt0 *= 2;
    if (VERBOSE) 
        fprintf(stdlog, "} set to %.2e ms (error=%.2e) \n", m_dt0, error);
}

//**************************************************************************
//                 A D A P T I V E   M E T H O D                 
//**************************************************************************
// Iterations begin with buffers update, so
// be sure to pass Ca field iterated over dt/2 !!
//**************************************************************************

void SimulationObj::Adaptive(double T)  {
  VectorObj Next(FieldObj::Size); 
  FieldObj  Next2(*Ca), oldCa(*Ca);
  BufferArray oldBuf(*Buffers);
  BufferArray BA(*Buffers); 

  double    one, two, buf_dt, oldDt;
  int       rvalue, total_steps = 0, between_divides = 0;
  double    error;
  double    dt_new;

  double dt = m_dt0; 

  if (m_ODEaccuracy > 1e-3) m_ODEaccuracy = 1e-3;
  double between_checks = m_n0;

  RunStatusString *ODEstatus = 0, *status = 0;

  if (VERBOSE > 2) status = new RunStatusString(30,"time",&(Ca->Time),T, "dt", &dt);
  else if (VERBOSE) status = new RunStatusString(1,"time",&(Ca->Time),T);	
  if (VERBOSE > 3) ODEstatus = status;

  while(1)
    {
    int i = 1;
    oldCa = *Ca;        oldBuf = *Buffers;   // store variables to recover 
    Gates->saveState(); oldDt = dt;        // when m_maxError is exceeded

    do         // loop over n="between_checks" iterations
      { 
      if ( ( dt_new = dt * m_dtStretch ) > m_dtMax ) dt_new = m_dtMax;
      buf_dt = 0.5 * ( dt + dt_new );
      Buffers->Step(BufStep, Next, *Ca, buf_dt, T); 
      Gates->RungeKuttaAdaptive(buf_dt, m_ODEaccuracy, 0, ODEstatus, T);
      dt = dt_new;                            
      rvalue = Ca->Step(CaStep, Next, *Buffers, dt, T);
      Plots->draw_all();
      total_steps ++; 
      between_divides ++; 

      if (VERBOSE > 1 && VERBOSE <= 7) status->update('.');
      else if (VERBOSE > 7) 
        fprintf(stdlog, "\n > Adaptive step #%d(of %d): T=%g(%g) dt=%.3g", 
              i, int(between_checks+0.5), Ca->Time, Gates->Time, dt);
      if ( (i++ == int(between_checks + 0.5)) && (Ca->Time + dt < T) ) break; 
      }
    while (!rvalue);

    Buffers->Step(BufStep, Next, *Ca, dt, T);
    Gates->RungeKuttaAdaptive(dt, m_ODEaccuracy, 0, ODEstatus, T);
    Plots->draw_all();
    if (rvalue) {                // procedure return point
      if (status) delete status;
      return;
    }

    Ca->Step(   CaStep, Next2, *Buffers, 0.75 * dt, -1);  // check accuracy
    BA = *Buffers;
    // BA.Step(BufStep, Next, Ca, 0.875 * dt, T);
    Next2.Step(CaStep, Next,  BA, 0.25 * dt, -1);
    two = Next.norm();
    Ca->Step(CaStep, Next, *Buffers, dt, -1);
    one = Next.norm();
    error = fabs( 2 * (two - one) / (two + one + 1.0e-12) );

    if (VERBOSE > 7) 
      fprintf(stdlog,"\n > accuracy check after %d steps: T=%g dt=%g error=%g ", 
              between_divides, Ca->Time, dt, error);
       
    if (error > m_maxError || isnan(error) ) {
      m_nStretch = 1 + (m_nStretch - 1) * m_reduce; 
      m_dtStretch = 1 + (m_dtStretch - 1) * m_reduce;
      if (m_n0 > 1) m_n0--;
      if (m_nMax > 1) m_nMax--;
      fprintf(stdlog,"\n*** Instability correction:  maxError exceeded (error = %.2g%% > %.2g%%)\n", error*100, m_maxError*100);
      fprintf(stdlog,"  > Stepping back by %.3gms and halving time step to %.3ems\n", Ca->Time-oldCa.Time, dt * 0.5);
      fprintf(stdlog,"  > Changing adaptive time-step parameters: \n");
      fprintf(stdlog,"  > dtStretch=%g, n0=%d, nStretch=%g, nMax=%d\n", m_dtStretch, m_n0, m_nStretch, m_nMax); 
      *Ca = oldCa; *Buffers = oldBuf; Gates->recoverState();
      Buffers->Step(BufStep, Next, *Ca, oldDt * 0.75, T); 
      Gates->RungeKuttaAdaptive(oldDt * 0.75, m_ODEaccuracy, 0, ODEstatus, T);
      dt *= 0.5;
      rvalue = Ca->Step(CaStep, Next, *Buffers, 0.5 * dt, T);
      Plots->draw_all();
      between_checks = m_n0;
      between_divides = 0;
      } 
    else if ( error > m_accuracy )  // halving the time step
       {
       Next2.Time = Ca->Time;
       *Ca = Next2;
       Ca->Time += 0.75 * dt;
       dt /= 2; 
       between_checks = m_n0;
       if (VERBOSE > 1 && VERBOSE <= 7)  status->update('|');
       else if (VERBOSE > 7) 
         fprintf(stdlog, "\n# dt halved @ %d steps: t=%gms, dt=%.2ems, error=%.2e ", 
                between_divides, Ca->Time, dt, error);
       between_divides = 0;
       }
     else   // accuracy within limits (error < eps), time step unchanged
       {
       if (VERBOSE > 1 && VERBOSE <= 7)  status->update('+');
       if ( int(between_checks *= m_nStretch) > m_nMax) between_checks = m_nMax;       
       *Ca = Next;
       Ca->Time += dt;
       }
     }
}

//**************************************************************************
//                                R U N                
//**************************************************************************

void SimulationObj::Run() {

  bool adaptive;
  double t, T = 0.0;
  Plots->draw_all();

  for (int i = 1; i <= Params->token_count("Run"); i++)  {
      
     getRun( *Params, i, &adaptive, m_PDEscheme, &t );
     int p = Params->token_index("current", i) + 1;
     if ( equal((*Params)[p], ASSIGN_TOKEN) ) p++;
     ExpressionObj Current(*Params, p, "Bad expression for the current", this, 0, &(Ca->Time), "t");

     Ca->current = &Current;
     T += t;
   
     char *method = StrCpy( getMethod(m_PDEscheme, CaStep, BufStep ) );

       if ( adaptive ) {
          getRun( *Params, i, 0, 0, 0, &m_dt0, &m_dtMax, &m_accuracy, &m_accuracy0, &m_maxError, 
                                    &m_dtStretch, &m_n0, &m_nStretch, &m_nMax, &m_ODEaccuracy);   
          if (VERBOSE) {
            fprintf(stdlog,"\n\n#### Running adaptive %s scheme for %g ms, current = ", method, t);
            Current.print(stdlog);  // make it crash here if expression for current is not valid
            fprintf(stdlog, "\n > dt0=%gms, dtMax=%gms, accuracy=%g, accuracy0=%g, maxError=%g dtStretch=%g\n",
                            m_dt0, m_dtMax, m_accuracy, m_accuracy0, m_maxError, m_dtStretch);
            fprintf(stdlog, " > n0 (initial step num)=%d, nStretch=%g, nMax=%d, ODEaccuracy=%g\n",
                            m_n0, m_nStretch, m_nMax, m_ODEaccuracy); 
	  }
          if (m_dt0 > m_dtMax) m_dt0 = m_dtMax;
          SetTimeStep();
          Adaptive(T);
          m_dt0 *= 10;
	}
        else {  // non-adaptive method
          getRun( *Params, i, 0, 0, 0, &m_dt0, &m_ODEaccuracy);
          int Tsteps = int(t / m_dt0 + 0.5);
          if (VERBOSE) {
            fprintf(stdlog,"\n#### Running non-adaptive %s scheme for %g ms, current = ", method, t);
            Current.print(stdlog); fflush(stdlog);
            Current.Evaluate();  // make it crash here is expression for current is not valid
            fprintf(stdlog,"[I]\n > time step = %g ms, number of iterations = %d\n", m_dt0, Tsteps);
          }
          FixedTimeStep(T, Tsteps);
	}
        Plots->redraw_all();
        delete [] method;

        if (VERBOSE) {
          double exper = calcium_gain(Ca, Buffers);
          double theor = *(this->ResolveID("Charge"));
          CHARGE_LOSS = ( exper == theor ) ? 0.0 : 200.0 * (theor - exper) / (exper + theor);
          fprintf(stdlog," ## charge loss = %.4g%% (net charge = %g, integrated flux = %g) \n", 
                  CHARGE_LOSS, exper, theor );
	}
        
  } 

  fprintf(stdlog, "\n### Simulation completed: total time = %g ms\n",T );
}
//**************************************************************************
//                           O D E   S O L V E R               
//**************************************************************************

ODESimulationObj::ODESimulationObj(TokenString &TS) : SimulationObj() {

     int pos;
     char temp[512];

     totalSimTime = get_sim_time(TS);
     Params = &TS;
     Gates   = new KineticObj(this);
     if ( TS.token_count("Import", &pos) ) Import( Params->line_string( pos + 1, temp ) );
     Plots   = new PlotArray(*this);
     Plots->redraw_all();

     m_dt0 = 1.0;
     m_ODEaccuracy = 1e-6;
     Params->get_param("ODE.dt0",0,&m_dt0);
     Params->get_param("ODE.accuracy",0,&m_ODEaccuracy);

  }                                 

//*******************************************************************************************

void ODESimulationObj::Run()  {
  double time = 0;

  RunStatusString *rss = 0;
  if (VERBOSE > 2) rss = new RunStatusString(30, "time", &(Gates->Time), 0);
  else if (VERBOSE) rss = new RunStatusString(1, "time", &(Gates->Time), 0);

  fprintf(stdlog,"\n\n#### Running the ODEs: total simulation time = %g ms\n", totalSimTime);

  for (int i = 1; i <= Params->token_count("Run"); i++) {
    
    getRun(*Params, i, 0, 0, &time, &m_dt0, &m_ODEaccuracy);

    fprintf(stdlog," ### ODE Run #%d: time = %g ms, accuracy = %g, initial time-step = %g ms\n", 
            i, time, m_ODEaccuracy, m_dt0);

    if (rss) rss->reset( Gates->Time + time );
    Gates->RungeKuttaAdaptive(time, m_ODEaccuracy, Plots, rss, totalSimTime, m_dt0);
    if (VERBOSE) fprintf(stdlog,"\n");
    Plots->redraw_all();
  }

  if (rss) delete rss;
}
//*******************************************************************************************************
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//*******************************************************************************************************
